from .memorix_base import MemorixBase
from .cache.cache_base import CacheBase as MemorixCacheBase
from .cache.cache_item import (
    CacheItem as MemorixCacheItem,
    CacheItemNoKey as MemorixCacheItemNoKey,
)
from .pubsub.pubsub_base import PubSubBase as MemorixPubSubBase
from .pubsub.pubsub_item import (
    PubSubItem as MemorixPubSubItem,
    PubSubItemNoKey as MemorixPubSubItemNoKey,
)
from .task.task_base import TaskBase as MemorixTaskBase
from .task.task_item import (
    TaskItem as MemorixTaskItem,
    TaskItemNoKey as MemorixTaskItemNoKey,
    TaskItemNoReturns as MemorixTaskItemNoReturns,
    TaskItemNoKeyNoReturns as MemorixTaskItemNoKeyNoReturns,
)
from dataclasses import dataclass



from redis import Redis


class RedisConnection(object):
    def __init__(
        self,
        redis_url: str,
    ) -> None:
        self.redis = Redis.from_url(redis_url)
        self.pubsub = self.redis.pubsub()



import os
from .example_schema_generated import Animal, Memorix, User
import multiprocessing
from time import sleep

redis_url = os.environ["REDIS_URL"]


def listen_to_message() -> None:
    memorix = Memorix(redis_url=redis_url)
    for res in memorix.pubsub.message.subscribe():
        print("message:", res.payload)


def listen_to_algo() -> None:
    memorix = Memorix(redis_url=redis_url)
    for res in memorix.task.runAlgo.dequeue():
        print("task:", res.payload)
        res.send_returns(returns=Animal.dog)


def start() -> None:
    memorix = Memorix(redis_url=redis_url)

    memorix.cache.user.set("uv", User(name="uv", age=29))

    user = memorix.cache.user.get("uv")
    if user is None:
        raise Exception("Didn't get user from redis")
    print(user.age)

    process = multiprocessing.Process(target=listen_to_message)
    process.start()

    for _num in (1, 2, 3, 4):
        sleep(0.1)
        res = memorix.pubsub.message.publish(payload="Heyy buddy")
        print("listeners:", res.subscribers_size)

    sleep(0.2)
    process.kill()

    task = multiprocessing.Process(target=listen_to_algo)
    task.start()

    for _num2 in (1, 2, 3, 4):
        sleep(0.1)
        queue = memorix.task.runAlgo.queue(payload="Im a task!")
        print("queue_size:", queue.queue_size)
        res2 = queue.get_returns()
        print("animal:", res2.value)

    sleep(0.2)
    task.kill()



import asyncio
import functools
from memorix_client_redis.hash_key import hash_key
from uuid import uuid4
from memorix_client_redis.json import from_json, to_json, bytes_to_str
import typing
from memorix_client_redis.memorix_base import MemorixBase
from .task_options import TaskDequequeOptions

KT = typing.TypeVar("KT")
PT = typing.TypeVar("PT")
RT = typing.TypeVar("RT")


class TaskItemQueue(object):
    def __init__(self, queue_size: int) -> None:
        self.queue_size = queue_size


class TaskItemQueueWithReturns(TaskItemQueue, typing.Generic[RT]):
    def __init__(
        self,
        queue_size: int,
        returns_id: str,
        returns_task: typing.Any,
    ) -> None:
        self.queue_size = queue_size
        self._returns_id = returns_id
        self._returns_task = returns_task

    def get_returns(self) -> RT:
        payload = None
        for message in self._returns_task.dequeue(key=self._returns_id):
            payload = message.payload
            break
        return typing.cast(RT, payload)

    async def async_get_returns(self) -> RT:
        payload = None
        async for message in self._returns_task.async_dequeue(key=self._returns_id):
            payload = message.payload
            break
        return typing.cast(RT, payload)


class TaskItemDequeue(typing.Generic[PT]):
    def __init__(
        self,
        payload: PT,
    ) -> None:
        self.payload = payload


class TaskItemDequeueWithReturns(TaskItemDequeue[PT]):
    def __init__(
        self,
        payload: PT,
        returns_id: str,
        returns_task: typing.Any,
    ) -> None:
        super().__init__(payload=payload)
        self._returns_id = returns_id
        self._returns_task = returns_task

    def send_returns(self, returns: RT) -> TaskItemQueue:
        return typing.cast(
            TaskItemQueue,
            self._returns_task.queue(key=self._returns_id, payload=returns),
        )

    async def async_send_returns(self, returns: RT) -> TaskItemQueue:
        return typing.cast(
            TaskItemQueue,
            self._returns_task.async_queue(key=self._returns_id, payload=returns),
        )


class TaskItem(typing.Generic[KT, PT, RT]):
    Options = TaskDequequeOptions

    def __init__(
        self,
        api: MemorixBase,
        id: str,
        payload_class: typing.Type[PT],
        returns_class: typing.Optional[typing.Type[RT]] = None,
        options: typing.Optional[TaskDequequeOptions] = None,
    ) -> None:
        self._api = api
        self._id = id
        self._payload_class = payload_class
        if returns_class is not None:
            self._returns_task = TaskItemNoReturns[str, RT](
                api=api,
                id="{0}_returns".format(id),
                payload_class=returns_class,
            )
        self._options = options

    def queue(self, key: KT, payload: PT) -> TaskItemQueueWithReturns[RT]:
        returns_id: str | None = None
        if hasattr(self, "_returns_task"):
            returns_id = str(uuid4())
            wrapped_payload_json = to_json(
                [returns_id, payload],
            )
        else:
            wrapped_payload_json = to_json(
                [payload],
            )

        queue_size = self._api._connection.redis.rpush(
            hash_key(api=self._api, id=self._id, key=key),
            wrapped_payload_json,
        )
        if returns_id is None:
            return typing.cast(
                TaskItemQueueWithReturns[RT],
                TaskItemQueue(queue_size=queue_size),
            )
        return TaskItemQueueWithReturns(
            queue_size=queue_size,
            returns_id=returns_id,
            returns_task=self._returns_task,
        )

    async def async_queue(self, key: KT, payload: PT) -> TaskItemQueueWithReturns[RT]:
        print("queue async")
        return typing.cast(TaskItemQueueWithReturns[RT], None)

    def dequeue(
        self,
        key: KT,
        options: typing.Optional[TaskDequequeOptions] = None,
    ) -> typing.Generator[TaskItemDequeueWithReturns[PT], None, None]:
        merged_options = self._options
        if self._api._default_options is not None:
            merged_options = TaskDequequeOptions.merge(
                self._api._default_options.task,
                self._options,
            )
        merged_options = TaskDequequeOptions.merge(
            merged_options,
            options,
        )

        while True:
            if merged_options is not None and merged_options.take_newest:
                [channel_bytes, data_bytes] = self._api._connection.redis.brpop(
                    hash_key(api=self._api, id=self._id, key=key),
                )
            else:
                [channel_bytes, data_bytes] = self._api._connection.redis.blpop(
                    hash_key(api=self._api, id=self._id, key=key),
                )

            data_str = bytes_to_str(data_bytes)

            if hasattr(self, "_returns_task"):
                seperator_index = data_str.index(",")
                returns_id = data_str[2 : seperator_index - 1]
                payload_str = data_str[seperator_index + 1 : -1]
                payload = from_json(
                    payload_str,
                    self._payload_class,
                )
                yield TaskItemDequeueWithReturns(
                    payload=payload,
                    returns_id=returns_id,
                    returns_task=self._returns_task,
                )
            else:
                payload_str = data_str[1:-1]
                payload = from_json(payload_str, self._payload_class)

                yield typing.cast(
                    TaskItemDequeueWithReturns[PT],
                    TaskItemDequeue(payload=payload),
                )

    async def async_dequeue(
        self,
        key: KT,
    ) -> typing.AsyncGenerator[TaskItemDequeueWithReturns[PT], None]:
        print("dequeue async")
        yield typing.cast(TaskItemDequeueWithReturns[PT], None)

    def clear(self, key: KT) -> None:
        self._api._connection.redis.delete(
            hash_key(api=self._api, id=self._id, key=key),
        )

    async def async_clear(self, key: KT) -> None:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            functools.partial(
                TaskItem.clear,
                self=self,
                key=key,
            ),
        )


class TaskItemNoKey(TaskItem[None, PT, RT]):
    # Different signature on purpose
    def queue(self, payload: PT) -> TaskItemQueueWithReturns[RT]:  # type: ignore
        return TaskItem.queue(self, key=None, payload=payload)

    # Different signature on purpose
    async def async_queue(self, payload: PT) -> TaskItemQueueWithReturns[RT]:  # type: ignore
        return await TaskItem.async_queue(self, key=None, payload=payload)

    # Different signature on purpose
    def dequeue(self) -> typing.Generator[TaskItemDequeueWithReturns[PT], None, None]:  # type: ignore
        return TaskItem.dequeue(self, key=None)

    # Different signature on purpose
    async def async_dequeue(self) -> typing.AsyncGenerator[TaskItemDequeueWithReturns[PT], None]:  # type: ignore
        return TaskItem.async_dequeue(self, key=None)

    # Different signature on purpose
    def clear(self) -> None:  # type: ignore
        TaskItem.clear(self, key=None)

    # Different signature on purpose
    async def async_clear(self) -> None:  # type: ignore
        await TaskItem.async_clear(self, key=None)


class TaskItemNoReturns(TaskItem[KT, PT, None]):
    def __init__(
        self,
        api: MemorixBase,
        id: str,
        payload_class: typing.Type[PT],
    ) -> None:
        super().__init__(
            api=api,
            id=id,
            payload_class=payload_class,
            returns_class=None,
        )

    # Different signature on purpose
    def queue(self, key: KT, payload: PT) -> TaskItemQueue:  # type: ignore
        return typing.cast(
            TaskItemQueue,
            TaskItem.queue(self, key=key, payload=payload),
        )

    # Different signature on purpose
    async def async_queue(self, key: KT, payload: PT) -> TaskItemQueue:  # type: ignore
        return await TaskItem.async_queue(self, key=key, payload=payload)

    # Different signature on purpose
    def dequeue(self, key: KT) -> typing.Generator[TaskItemDequeue[PT], None, None]:  # type: ignore
        return TaskItem.dequeue(self, key=key)

    # Different signature on purpose
    async def async_dequeue(self, key: KT) -> typing.AsyncGenerator[TaskItemDequeue[PT], None]:  # type: ignore
        return TaskItem.async_dequeue(self, key=key)

    # Different signature on purpose
    def clear(self, key: KT) -> None:
        TaskItem.clear(self, key=key)

    # Different signature on purpose
    async def async_clear(self, key: KT) -> None:
        await TaskItem.async_clear(self, key=key)


class TaskItemNoKeyNoReturns(TaskItemNoReturns[None, PT]):
    # Different signature on purpose
    def queue(self, payload: PT) -> TaskItemQueue:  # type: ignore
        return typing.cast(
            TaskItemQueue,
            TaskItem.queue(self, key=None, payload=payload),
        )

    # Different signature on purpose
    async def async_queue(self, payload: PT) -> TaskItemQueue:  # type: ignore
        return await TaskItem.async_queue(self, key=None, payload=payload)

    # Different signature on purpose
    def dequeue(self) -> typing.Generator[TaskItemDequeue[PT], None, None]:  # type: ignore
        return TaskItem.dequeue(self, key=None)

    # Different signature on purpose
    async def async_dequeue(self) -> typing.AsyncGenerator[TaskItemDequeue[PT], None]:  # type: ignore
        return TaskItem.async_dequeue(self, key=None)

    # Different signature on purpose
    def clear(self) -> None:  # type: ignore
        TaskItem.clear(self, key=None)

    # Different signature on purpose
    async def async_clear(self) -> None:  # type: ignore
        await TaskItem.async_clear(self, key=None)



import asyncio
import functools
from memorix_client_redis.hash_key import hash_key
from memorix_client_redis.json import from_json, to_json, bytes_to_str
from memorix_client_redis.memorix_base import MemorixBase
from typing import AsyncGenerator, Dict, Generator, Generic, Type, TypeVar, Union, cast

KT = TypeVar("KT")
PT = TypeVar("PT")


class PubSubItemPublish(object):
    def __init__(
        self,
        subscribers_size: int,
    ) -> None:
        self.subscribers_size = subscribers_size


class PubSubItemSubscribe(Generic[PT]):
    def __init__(
        self,
        payload: PT,
    ) -> None:
        self.payload = payload


class PubSubItem(Generic[KT, PT]):
    def __init__(
        self,
        api: MemorixBase,
        id: str,
        payload_class: Type[PT],
    ) -> None:
        self._api = api
        self._id = id
        self._payload_class = payload_class

    def publish(self, key: KT, payload: PT) -> PubSubItemPublish:
        payload_json = to_json(payload)
        subscribers_size = self._api._connection.redis.publish(
            hash_key(api=self._api, id=self._id, key=key),
            payload_json,
        )
        return PubSubItemPublish(subscribers_size=subscribers_size)

    async def async_publish(self, key: KT, payload: PT) -> PubSubItemPublish:
        loop = asyncio.get_running_loop()
        res = await loop.run_in_executor(
            None,
            functools.partial(
                PubSubItem.publish,
                self=self,
                key=key,
                payload=payload,
            ),
        )
        return res

    def subscribe(self, key: KT) -> Generator[PubSubItemSubscribe[PT], None, None]:
        sub = self._api._connection.redis.pubsub()
        sub.subscribe(hash_key(api=self._api, id=self._id, key=key))
        for message in cast(
            Generator[Dict[str, Union[int, bytes]], None, None],
            sub.listen(),  # type: ignore
        ):
            data_bytes = message["data"]
            if isinstance(data_bytes, bytes):
                data_str = bytes_to_str(data_bytes)
                data = from_json(value=data_str, data_class=self._payload_class)
                yield PubSubItemSubscribe(payload=data)

    async def async_subscribe(
        self,
        key: KT,
    ) -> AsyncGenerator[PubSubItemSubscribe[PT], None]:
        sub = self._api._connection.redis.pubsub()
        sub.subscribe(hash_key(api=self._api, id=self._id, key=key))
        loop = asyncio.get_running_loop()
        while True:
            message = await loop.run_in_executor(
                None,
                functools.partial(
                    sub.get_message,
                    ignore_subscribe_messages=True,
                ),
            )
            if message is not None:
                data_bytes = message["data"]
                if isinstance(data_bytes, bytes):
                    data_str = bytes_to_str(data_bytes)
                    data = from_json(value=data_str, data_class=self._payload_class)
                    yield PubSubItemSubscribe(payload=data)


class PubSubItemNoKey(PubSubItem[None, PT]):
    # Different signature on purpose
    def publish(self, payload: PT) -> PubSubItemPublish:  # type: ignore
        return PubSubItem.publish(self, key=None, payload=payload)

    # Different signature on purpose
    async def async_publish(self, payload: PT) -> PubSubItemPublish:  # type: ignore
        return await PubSubItem.async_publish(self, key=None, payload=payload)

    # Different signature on purpose
    def subscribe(self) -> Generator[PubSubItemSubscribe[PT], None, None]:  # type: ignore
        return PubSubItem.subscribe(self, key=None)

    # Different signature on purpose
    def async_subscribe(self) -> AsyncGenerator[PubSubItemSubscribe[PT], None]:  # type: ignore
        return PubSubItem.async_subscribe(self, key=None)



# flake8: noqa
import typing

if typing.TYPE_CHECKING:
    from dataclasses import dataclass
else:
    from memorix_client_redis import dataclass

from enum import Enum
from memorix_client_redis import (
    MemorixBase,
    MemorixCacheBase,
    MemorixCacheItem,
    MemorixCacheItemNoKey,
    MemorixPubSubBase,
    MemorixPubSubItem,
    MemorixPubSubItemNoKey,
    MemorixTaskBase,
    MemorixTaskItem,
    MemorixTaskItemNoKey,
    MemorixTaskItemNoReturns,
    MemorixTaskItemNoKeyNoReturns,
)


class Animal(str, Enum):
    dog = "dog"
    cat = "cat"
    person = "person"


@dataclass
class User(object):
    name: str
    age: typing.Optional[int]


class MemorixCache(MemorixCacheBase):
    def __init__(self, api: MemorixBase) -> None:
        super().__init__(api=api)

        self.favoriteAnimal = MemorixCacheItem[str, "Animal"](
            api=api,
            id="favoriteAnimal",
            payload_class=Animal,
        )
        self.user = MemorixCacheItem[str, "User"](
            api=api,
            id="user",
            payload_class=User,
        )


class MemorixPubSub(MemorixPubSubBase):
    def __init__(self, api: MemorixBase) -> None:
        super().__init__(api=api)

        self.message = MemorixPubSubItemNoKey[str](
            api=api,
            id="message",
            payload_class=str,
        )


class MemorixTask(MemorixTaskBase):
    def __init__(self, api: MemorixBase) -> None:
        super().__init__(api=api)

        self.runAlgo = MemorixTaskItemNoKey[str, "Animal"](
            api=api,
            id="runAlgo",
            payload_class=str,
            returns_class=Animal,
        )


class Memorix(MemorixBase):
    def __init__(
        self,
        redis_url: str,
        ref: typing.Optional[MemorixBase] = None,
    ) -> None:
        super().__init__(redis_url=redis_url, ref=ref)

        self._namespace_name_tree = []

        self.cache = MemorixCache(self)
        self.pubsub = MemorixPubSub(self)
        self.task = MemorixTask(self)
